"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.multicastStream = exports.cloneTarStream = exports.pipePromise = exports.drainStream = exports.streamToBuffer = exports.normalizeTarEntry = void 0;
const path = require("path");
const tar = require("tar-stream");
function noop() {
}
function normalizeTarEntry(name) {
    if (!name) {
        return '';
    }
    return path.posix.normalize(name).replace(/^\/+|\/+$/g, '') || '.';
}
exports.normalizeTarEntry = normalizeTarEntry;
async function streamToBuffer(stream) {
    let onData;
    let onError;
    let onEnd;
    try {
        const buffers = await new Promise((resolve, reject) => {
            const chunks = [];
            stream.on('end', (onEnd = () => resolve(chunks)));
            stream.on('error', (onError = reject));
            stream.on('data', (onData = (data) => chunks.push(data)));
        });
        return Buffer.concat(buffers);
    }
    finally {
        stream.removeListener('data', onData ?? noop);
        stream.removeListener('error', onError ?? noop);
        stream.removeListener('end', onEnd ?? noop);
    }
}
exports.streamToBuffer = streamToBuffer;
async function drainStream(stream) {
    let onError;
    let onEnd;
    try {
        return await new Promise((resolve, reject) => {
            stream.on('error', (onError = reject));
            stream.on('end', (onEnd = resolve));
            stream.resume();
        });
    }
    finally {
        stream.removeListener('error', onError ?? noop);
        stream.removeListener('end', onEnd ?? noop);
    }
}
exports.drainStream = drainStream;
async function pipePromise(pipeFrom, pipeTo) {
    let onError;
    let onFinish;
    try {
        return await new Promise((resolve, reject) => {
            pipeFrom.on('error', (onError = reject));
            pipeTo.on('error', onError);
            pipeTo.on('finish', (onFinish = () => resolve(pipeTo)));
            pipeFrom.pipe(pipeTo);
        });
    }
    finally {
        pipeFrom.removeListener('error', onError ?? noop);
        pipeTo.removeListener('error', onError ?? noop);
        pipeTo.removeListener('finish', onFinish ?? noop);
    }
}
exports.pipePromise = pipePromise;
async function cloneTarStream(sourceTarStream, opts) {
    const extract = tar.extract();
    const pack = tar.pack();
    const origPush = pack.push;
    pack.push = function () {
        origPush.apply(this, arguments);
        return true;
    };
    let packOnError;
    let sourceTarStreamOnError;
    try {
        return await new Promise((resolve, reject) => {
            packOnError = reject;
            sourceTarStreamOnError = reject;
            sourceTarStream.on('error', sourceTarStreamOnError);
            pack.on('error', packOnError);
            extract.on('error', reject);
            extract.on('entry', async (header, stream, callback) => {
                if (opts?.onEntry) {
                    try {
                        await opts.onEntry(pack, header, stream);
                        callback();
                    }
                    catch (err) {
                        callback(err);
                    }
                }
                else {
                    const buf = await streamToBuffer(stream);
                    pack.entry(header, buf, callback);
                }
            });
            extract.once('finish', async () => {
                try {
                    if (opts?.onFinish) {
                        await opts.onFinish(pack);
                    }
                    pack.finalize();
                    resolve(pack);
                }
                catch (err) {
                    reject(err);
                }
            });
            sourceTarStream.pipe(extract);
        });
    }
    finally {
        sourceTarStream.removeListener('error', sourceTarStreamOnError ?? noop);
        pack.removeListener('error', packOnError ?? noop);
    }
}
exports.cloneTarStream = cloneTarStream;
async function multicastStream(fromStream, toStreams) {
    if (toStreams.length === 0) {
        return;
    }
    let onError;
    try {
        return await new Promise((resolve, reject) => {
            Promise.all(toStreams.map(async (toStream) => {
                let onFinish;
                try {
                    return await new Promise((toStreamResolve) => toStream.on('finish', (onFinish = toStreamResolve)));
                }
                finally {
                    toStream.removeListener('finish', onFinish ?? noop);
                }
            }))
                .then(() => resolve())
                .catch(reject);
            fromStream.on('error', (onError = reject));
            for (const toStream of toStreams) {
                toStream.on('error', onError);
                fromStream.pipe(toStream);
            }
        });
    }
    finally {
        fromStream.removeListener('error', onError ?? noop);
        toStreams.forEach((toStream) => toStream.removeListener('error', onError ?? noop));
    }
}
exports.multicastStream = multicastStream;
//# sourceMappingURL=index.js.map